/** EMULib Emulation Library *********************************/
/**                                                         **/
/**                          C24XX.c                        **/
/**                                                         **/
/** This file contains emulation for the 24cXX series of    **/
/** serial EEPROMs. See C24XX.h for declarations.           **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 2017-2018                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/

#include "C24XX.h"
#include <stdio.h>

/** Internal Chip States *************************************/
#define RECV_CMD  0
#define RECV_ADDR 1
#define RECV_ADR2 2
#define RECV_DATA 3
#define SEND_DATA 4

/** Size24XX *************************************************/
/** Return the size of given chip model in bytes.           **/
/*************************************************************/
unsigned int Size24XX(C24XX *D)
{
  unsigned int Chip = D->Flags&C24XX_CHIP;
  return(Chip<=C24XX_24C1024? (0x80<<Chip):0);
}

/** Reset24XX ************************************************/
/** Reset the 24xx chip.                                    **/
/*************************************************************/
void Reset24XX(C24XX *D,byte *Data,unsigned int Flags)
{
  D->Pins  = C24XX_SDA|C24XX_SCL;
  D->Out   = C24XX_SDA|C24XX_SCL;
  D->State = RECV_CMD;
  D->Bits  = 0x0001;
  D->Cmd   = 0x00;
  D->Addr  = 0;
  D->Data  = Data;
  D->Flags = Flags;
}

/** Read24XX *************************************************/
/** Read value from the 24xx chip. Only bits 0,1 are used   **/
/** (see #defines).                                         **/
/*************************************************************/
byte Read24XX(C24XX *D)
{
//  if(D->Flags&C24XX_DEBUG)
//    printf("EEPROM SDA=%d, SCL=%d\n",D->Pins&C24XX_SDA? 1:0,D->Pins&C24XX_SCL? 1:0);

  return(D->Out);
}

/** Write24XX ************************************************/
/** Write value V into the 24xx chip. Only bits 0,1 are     **/
/** used (see #defines).                                    **/
/*************************************************************/
byte Write24XX(C24XX *D,byte V)
{
  static const unsigned int PageSize[16] =
  { 8,8,16,16,16,32,32,64,128,256,256,256,256,256,256,256 };

  word J;

//  if((D->Flags&C24XX_DEBUG)&&((D->Pins^V)&(C24XX_SDA|C24XX_SCL)))
//    printf("EEPROM SDA: %d=>%d, SCL: %d=>%d\n",
//      D->Pins&C24XX_SDA? 1:0,V&C24XX_SDA? 1:0,
//      D->Pins&C24XX_SCL? 1:0,V&C24XX_SCL? 1:0
//    );

  /* When SDA line changes while SCL=1... */
  if(((D->Pins^V)&C24XX_SDA) && (D->Pins&V&C24XX_SCL))
  {
    if(V&C24XX_SDA)
    {
      /* SDA=1: STOP condition */
      if(D->Flags&C24XX_DEBUG) printf("EEPROM STOP\n");
      D->State = RECV_CMD;
      D->Bits  = 0x0001;
    }
    else
    {
      /* SDA=0: START condition */
      if(D->Flags&C24XX_DEBUG) printf("EEPROM START\n");
      D->State = RECV_CMD;
      D->Bits  = 0x0001;
    }
  }
  /* When SCL line goes up... */
  else if((D->Pins^V)&V&C24XX_SCL)
  {
    if((D->State==SEND_DATA) && (D->Bits!=0x8000))
    {
      /* Report current output bit */
      D->Out = (D->Out&~C24XX_SDA)|(D->Bits&0x8000? C24XX_SDA:0);
      /* Shift output bits */
      D->Bits<<=1;
    }
    else if((D->State!=SEND_DATA) && (D->Bits<0x0100))
    {
      /* Shift input bits */
      D->Bits = (D->Bits<<1)|(V&C24XX_SDA? 1:0);
    }
    else
    {
      /* Depending on the state... */
      switch(D->State)
      {
        case RECV_CMD:
          D->Cmd   = D->Bits&0x00FF;
          D->State = (D->Cmd&0xF0)!=0xA0? RECV_CMD:D->Cmd&0x01? SEND_DATA:RECV_ADDR;
          if(D->Flags&C24XX_DEBUG) printf("EEPROM CMD=%02Xh(, ADDR=%Xh)\n",D->Cmd,D->Addr);
          break;
        case RECV_ADDR:
          D->Addr  = ((unsigned int)(D->Cmd&0x0E)<<7)+(D->Bits&0x00FF);
          D->Addr &= (0x80<<(D->Flags&C24XX_CHIP))-1;
          D->State = (D->Flags&C24XX_CHIP)>=C24XX_24C32? RECV_ADR2:RECV_DATA;
          if(D->Flags&C24XX_DEBUG) printf("EEPROM CMD=%02Xh, ADDR=%Xh\n",D->Cmd,D->Addr);
          break;
        case RECV_ADR2:
          D->Addr  = (D->Addr<<8)+(D->Bits&0x00FF);
          D->Addr &= (0x80<<(D->Flags&C24XX_CHIP))-1;
          D->State = RECV_DATA;
          if(D->Flags&C24XX_DEBUG) printf("EEPROM CMD=%02Xh, ADDR=%Xh\n",D->Cmd,D->Addr);
          break;
        case RECV_DATA:
          if(D->Flags&C24XX_DEBUG) printf("EEPROM WRITE[%Xh] <= %02Xh\n",D->Addr,D->Bits&0xFF);
          /* Write byte into EEPROM */
          D->Data[D->Addr] = D->Bits&0x00FF;
          /* Go to the next address inside N-byte page */
          J = PageSize[D->Flags&C24XX_CHIP]-1;
          D->Addr = ((D->Addr+1)&J)|(D->Addr&~J);
          break;
        case SEND_DATA:
          /* See below */
          break;
        default:
          D->State = RECV_CMD;
          break;
      }
      /* Acknowledge received byte, clear bit buffer */
      D->Bits = 0x0001;
      D->Out  = C24XX_SCL;
      /* If sending the next byte... */
      if(D->State==SEND_DATA)
      {
        /* Read byte from EEPROM */
        D->Bits = ((word)D->Data[D->Addr]<<8)|0x0080;
        if(D->Flags&C24XX_DEBUG) printf("EEPROM READ[%Xh] => %02Xh\n",D->Addr,D->Bits>>8);
        /* Go to the next address inside N-byte page */
        J = PageSize[D->Flags&C24XX_CHIP]-1;
        D->Addr = ((D->Addr+1)&J)|(D->Addr&~J);
      }
    }
  }

  /* New pin values */
  D->Pins = V;

  /* Done */
  return(D->Out);
}
